//! Rocks DB implementation of the storage layer
use crate::RocksDb;
use futures::{Future, Stream};
use nostr_rs_storage_base::{
    cursor::{check_future_call, FutureValue},
    Error, EventFilter, Storage,
};
use nostr_rs_types::types::Event;
use rocksdb::{BoundColumnFamily, DBIteratorWithThreadMode, DB};
use std::{
    collections::VecDeque,
    pin::Pin,
    sync::Arc,
    task::{Context, Poll},
};

type CurrentEventByPrefixFuture<'a> = Pin<
    Box<
        dyn Future<
                Output = Result<Option<nostr_rs_types::types::Event>, nostr_rs_storage_base::Error>,
            > + Send
            + 'a,
    >,
>;

pub struct Cursor<'a> {
    /// Reference to the rocks db database. This is useful to load the event
    /// data, because in the secondary indexes we only store the event ID. It
    /// could be possible to avoid a reference and speed up things even more to
    /// duplicate the event, but that would mean it would use way more disk
    /// space.
    pub db: &'a RocksDb,
    /// List of filters to apply to the events, before returning. If no filter
    /// is given each events from the secondary index will be returned,
    /// otherwise the events will be filtered by the given filter, and only
    /// those events that comply will be returned
    pub filter: Option<EventFilter>,
    /// Reference to the namespace to use to query the secondary index. If none
    /// is given the secondary_index_iterator must be constructed outside this
    /// wrapper.
    pub namespace: Option<Arc<BoundColumnFamily<'a>>>,
    /// The current secondary index iterator. If none is given the iterator will
    /// try to create one using the namespace property and the first prefix from
    /// prefixes (it will also be copied to current_prefix)
    pub secondary_index_iterator: Option<DBIteratorWithThreadMode<'a, DB>>,
    pub current_prefix: Vec<u8>,
    pub prefixes: VecDeque<Vec<u8>>,
    pub limit: Option<usize>,
    pub returned: usize,

    pub future_event: Option<CurrentEventByPrefixFuture<'a>>,
}

impl<'a> Cursor<'a> {
    /// Selects the next prefix available and starts an iterator using the
    /// secondary index. If no prefix is available from prefixes the functions
    /// return None, signalling upstream the are no more results
    fn select_next_prefix_using_secondary_index(&mut self) -> Option<()> {
        self.secondary_index_iterator = None;
        let prefix = self.prefixes.pop_front()?;
        self.secondary_index_iterator = Some(
            self.db
                .db
                .prefix_iterator_cf(self.namespace.as_ref()?, prefix.clone()),
        );
        self.current_prefix = prefix;
        Some(())
    }
}

impl<'a> Stream for Cursor<'a> {
    type Item = Result<Event, Error>;

    fn size_hint(&self) -> (usize, Option<usize>) {
        (0, None)
    }

    fn poll_next(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Option<Self::Item>> {
        if Some(self.returned) == self.limit {
            return Poll::Ready(None);
        }

        let this = Pin::into_inner(self);
        let db = this.db;

        match check_future_call(&mut this.future_event, &this.filter, cx) {
            FutureValue::Found(event) => {
                this.returned += 1;
                return Poll::Ready(Some(event));
            }
            FutureValue::Pending => return Poll::Pending,
            FutureValue::FoundNotMatch | FutureValue::Ended => {}
        }

        loop {
            let secondary_index = if let Some(iterator) = this.secondary_index_iterator.as_mut() {
                iterator
            } else {
                if this.namespace.is_some() {
                    let _ = this.select_next_prefix_using_secondary_index();
                    if let Some(iterator) = this.secondary_index_iterator.as_mut() {
                        iterator
                    } else {
                        return Poll::Ready(None);
                    }
                } else {
                    // No secondary index is used to query, this means the query is
                    // using the ID filter, so it is more efficient to use the
                    // primary index to prefetch events that may satisfy the query
                    return if let Some(prefix) = this.prefixes.pop_front() {
                        this.future_event = Some(db.get_event(prefix));
                        match check_future_call(&mut this.future_event, &this.filter, cx) {
                            FutureValue::Found(event) => {
                                this.returned += 1;
                                Poll::Ready(Some(event))
                            }
                            FutureValue::Pending => Poll::Pending,
                            FutureValue::FoundNotMatch | FutureValue::Ended => continue,
                        }
                    } else {
                        Poll::Ready(None)
                    };
                }
            };

            return match secondary_index.next() {
                Some(Ok((key, value))) => {
                    if !key.starts_with(&this.current_prefix) {
                        let _ = this.select_next_prefix_using_secondary_index();
                        continue;
                    }

                    this.future_event = Some(db.get_event(value));
                    match check_future_call(&mut this.future_event, &this.filter, cx) {
                        FutureValue::Found(event) => {
                            this.returned += 1;
                            Poll::Ready(Some(event))
                        }
                        FutureValue::Pending => Poll::Pending,
                        FutureValue::FoundNotMatch | FutureValue::Ended => continue,
                    }
                }
                Some(Err(err)) => {
                    this.returned += 1;
                    Poll::Ready(Some(Err(Error::Internal(err.to_string()))))
                }
                None => {
                    let _ = this.select_next_prefix_using_secondary_index();
                    continue;
                }
            };
        }
    }
}
